#!/usr/bin/perl -w

use strict;
use Digest::MD5 qw(md5_base64);
use CGI qw(header *html *div font table caption TR th td a hr cookie remote_addr user_agent);
use DBI;
require './config.ph';

if (!$::DSN) {
	print
	header (-charset => 'koi8-r'),
	start_html,
	font ({-size => '+1', -color => 'red'}, '     .   config.ph'),
	end_html;
	exit;
}

if (grep (defined $_, map cookie ($_), qw(aname auth)) != 2 ) {	# Check cookie
	print header (-location => 'login.cgi');
	exit;
}

my ($err, $rights);
my $dbh = DBI->connect ($::DSN, $::DBUser, $::DBPass, {ShowErrorStatement => 1, AutoCommit => 0});
if (!$dbh) {
	$err = '    : ' . DBI->errstr;
} else {
	my $u = $dbh->selectrow_arrayref ("SELECT aname, passwd, cookie, type FROM admins WHERE aname = " . $dbh->quote (cookie ('aname')));
	if ($dbh->errstr) {
		$err = '    : ' . $dbh->errstr;
	} elsif ($u && cookie ('auth') eq md5_base64 ($u->[2], remote_addr, user_agent, @$u [0, 1])) {
		$rights = $u->[3];
	} else {
		if ($u) {
			$dbh->do ("UPDATE admins SET cookie = '' WHERE aname = " . $dbh->quote ($u->[0]));
			$dbh->commit;
		}
		print header (-location => 'login.cgi');
		exit;
	}
}

print
header (-charset => 'koi8-r'),
start_html (-title => ' ', -encoding => 'koi8-r'),
start_div ({-align => 'center'});
END {$dbh->disconnect if $dbh; print $err ? font ({-size => '+1', -color => 'red'}, $err) : '', end_div, end_html}

exit if $err;

my ($admins, $roles, $users, $res) = ([], [], []);

if ($rights eq 'g') {
	$res = $dbh->selectall_arrayref ("SELECT aname FROM admins WHERE type = 'c' ORDER BY aname");
	if ($dbh->errstr) {
		$err = '    : ' . $dbh->errstr;
		exit;
	}
	push @$admins, $_->[0] for @$res;
}

$res = $dbh->selectall_arrayref ("SELECT r1.rolid, r1.type, r1.rname, r2.rname FROM roles r1, roles r2, hierarchy h WHERE r1.rolid = h.child AND h.parent = r2.rolid ORDER BY r1.rolid, r2.rolid");
if ($dbh->errstr) {
	$err = '    : ' . $dbh->errstr;
	exit;
}
for (@$res) {
	@{$roles->[$_->[0]]} {qw(type name)} = @$_ [1, 2] if !defined $roles->[$_->[0]];
	push @{$roles->[$_->[0]]->{parent}}, $_->[3] if $_->[2] ne $_->[3];
}

$res = $dbh->selectall_arrayref ("SELECT u.uname, u.addr, u.mask, r.rname, u.useid FROM users u, roles r WHERE u.rolid = r.rolid ORDER BY u.uname, u.mask DESC, u.addr");
if ($dbh->errstr) {
	$err = '    : ' . $dbh->errstr;
	exit;
}
push @$res, shift @$res if !length $res->[0]->[0];
for (@$res) {
	my $h = {name => $_->[0] || '*', role => $_->[3], id => $_->[4]};
#	print STDERR "addr: $_->[1], mask: $_->[2]\n";
	if (!$_->[2]) {
		$h->{host} = '*';
	} else {
		$h->{host} = join '.', unpack "C4", pack "N", $_->[1];
		my ($bits, $mask);
		for ($bits = 32, $mask = 0xffffffff; $bits >= 0 && $_->[2] != $mask; $bits--, $mask <<= 1)
			{}
		if ($bits < 0) {
			$h->{host} .= '/' . join '.', unpack "C4", pack "N", $_->[2];
		} elsif ($bits != 32) {
			$h->{host} .= "/$bits";
		}
	}
	push @$users, $h;
}

print
table ({-border => 1}, caption (font ({-size => '+1'}, ' ')),
	TR (th ([qw( )])),
	TR (
		td (cookie ('aname')),
		td (a ({-href => "admin.cgi?action=passwd&value=" . cookie ('aname')}, ' '))
	),
	map (TR (
		td ($_),
		td (
			a ({-href => "admin.cgi?action=passwd&value=$_"}, ' '),
			a ({-href => "admin.cgi?action=delete&value=$_"}, '')
		)
	), @$admins),
	$rights eq 'g' ? TR (td ({-colspan => 2, align => 'center'}, a ({-href => "admin.cgi?action=new"}, '  '))) : ''
), hr,
table ({-border => 1}, caption (font ({-size => '+1'}, ' ')),
	TR (th ([qw(   )])),
	map (TR (
		td ($_->{type} eq 'g' ? 'General' : 'Custom'),
		td ($_->{name}),
		td ($_->{parent} ? "@{$_->{parent}}" : '&nbsp'),
		td (
			a ({-href => "role.cgi?action=view&value=$_->{name}"}, ''),
			$_->{type} eq $rights ? (
				a ({-href => "role.cgi?action=edit&value=$_->{name}"}, ''),
				a ({-href => "role.cgi?action=delete&value=$_->{name}"}, ''),
			) : '',
		)
	), grep defined $_, @$roles),
	TR (td ({-colspan => 4, align => 'center'}, a ({-href => "role.cgi?action=new"}, '  ')))
), hr,
table ({-border => 1}, caption (font ({-size => '+1'}, '  ')),
	TR (
		th (''),
		th ('',),
		th (''),
		$rights eq 'c' ? th ('') : ''
	),
	map (TR (
		td ($_->{name}),
		td ($_->{host}),
		td ($_->{role}),
		$rights eq 'c'
		? td (
			a ({-href => "user.cgi?action=edit&value=$_->{id}"}, ''),
			a ({-href => "user.cgi?action=delete&value=$_->{id}"}, '')
		) : ''
	), @$users),
	$rights	eq 'c'
	? TR (td ({-colspan => 4, -align => 'center'}, a ({-href => 'user.cgi?action=new'}, '  ')))
	: '',
);
