#!/usr/bin/perl -w

use strict;
use Digest::MD5 qw(md5_base64);
use CGI qw(header *html *div font table caption TR th td a b br hr cookie remote_addr user_agent param url *form textfield checkbox_group checkbox button submit hidden);
use DBI;
require './config.ph';

if (!$::DSN) {
	print
	header (-charset => 'koi8-r'),
	start_html,
	font ({-size => '+1', -color => 'red'}, '     .   config.ph'),
	end_html;
	exit;
}

use constant	RoleNamePattern => '^\w+$';
use constant	URLPattern => '^\w+://([^/]+\.)?\w+\.\w+($|/)';

if (grep (defined $_, map cookie ($_), qw(aname auth)) != 2 ) {	# Check cookie
	print header (-location => 'login.cgi');
	exit;
}

my ($err, $rights, $res);
my $dbh = DBI->connect ($::DSN, $::DBUser, $::DBPass, {ShowErrorStatement => 1, AutoCommit => 0});
if (!$dbh) {
	$err = '    : ' . DBI->errstr;
} else {
	my $u = $dbh->selectrow_arrayref ("SELECT passwd, cookie, type FROM admins WHERE aname = " . $dbh->quote (cookie ('aname')));
	if ($dbh->errstr) {
		$err = '    : ' . $dbh->errstr;
	} elsif ($u && cookie ('auth') eq md5_base64 ($u->[1], remote_addr, user_agent, cookie ('aname'), $u->[0])) {
		$rights = $u->[2];
	} else {
		if ($u) {
			$dbh->do ("UPDATE admins SET cookie = '' WHERE aname = " . $dbh->quote (cookie ('aname')));
			$dbh->commit;
		}
		print header (-location => 'login.cgi');
		exit;
	}
}

print
header (-charset => 'koi8-r'),
start_html (-title => ' ', -encoding => 'koi8-r'),
start_div ({-align => 'center'});
END {
	$dbh->disconnect if $dbh;
	print
	$err ? font ({-size => '+1', -color => 'red'}, $err) : '',
	hr,
	a ({-href => 'main.cgi'}, '   '),
	end_div,
	end_html;
}

exit if $err;

if (!defined param ('action') || !grep param ('action') eq $_, qw(new view edit delete)) {
	$err = 'oe ';
	exit;
}

if (param ('action') eq 'new') {
	$res = $dbh->selectall_arrayref ("SELECT rolid, rname FROM roles" . ($rights eq 'g' ? " WHERE type = 'g'" : ''));
	if ($dbh->errstr) {
		$err = '    : ' . $dbh->errstr;
		exit;
	}
	my $rolids = {map @$_, @$res};
	if (param ('save')) {
		if (!defined param ('rname')) {
			$err = '   ';
		} elsif (param ('rname') !~ /@{[RoleNamePattern]}/) {
			$err = '  ';
		} elsif ($rights ne 'g' && !param ('parents')) {
			$err = '   ';
		} elsif (grep !$rolids->{$_}, param ('parents')) {
			$err = '  ';
		} else {
			$res = $dbh->do ("INSERT INTO roles (rname, type) VALUES (" . $dbh->quote (param ('rname')) . ",'".$rights."')");
			if ($dbh->errstr) {
				$err = '    : ' . $dbh->errstr;
			} else {
				$res = $dbh->selectcol_arrayref ("SELECT rolid FROM roles WHERE rname = " . $dbh->quote (param ('rname')));
				if ($dbh->errstr) {
					$err = '    : ' . $dbh->errstr;
				} elsif (!$res) {
					$err = '  ';
				} else {
					for ($res->[0], param ('parents')) {
						$dbh->do ("INSERT INTO hierarchy VALUES ($_, $res->[0])");
						if ($dbh->errstr) {
							$err = '    : ' . $dbh->errstr;
							last;
						}
					}
				}
			}
		}
	}
	if (param ('save') && !$err) {
		$dbh->commit;
		print font ({-size => '+1'}, '  .', a ({-href => url (-relative => 1) . '?action=edit&value=' . param ('rname')}, ' '), '  ');
	} else {
		print font ({-size => '+1', -color => 'red'}, $err), hr if $err;
		undef $err;
		print
		start_form ('post', url (-relative => 1)),
		table ({-border => 1},
			caption (font ({-size => '+1'}, '  ')),
			TR (th (' '), td (textfield ('rname', undef, 20, 20))),
			TR (th (''), td (checkbox_group ('parents', [sort {$a <=> $b} keys %$rolids], undef, undef, $rolids, undef, undef, 3))),
			TR (th ({-colspan => 2, align => 'center'}, submit ('save', '')))
		),
		hidden ('action'),
		end_form;
	}
} else {
	if (!defined param ('value')) {
		$err = '   ';
		exit;
	}

	if (param ('value') !~ /@{[RoleNamePattern]}/) {
		$err = '  ';
		exit;
	}

	$res = $dbh->selectrow_arrayref ("SELECT rolid, type FROM roles WHERE rname = " . $dbh->quote (param ('value')));
	if ($dbh->errstr) {
		$err = '    : ' . $dbh->errstr;
		exit;
	}

	if (!$res) {
		$err = '  ';
		exit;
	}

	my ($rolid, $roltype) = @$res;
	if (param ('action') eq 'view') {
		my ($lists, $perms) = ([[], []], [[], [], []]);

		# Read lists
		$res = $dbh->selectall_arrayref ("SELECT l.action, l.url FROM lists l, hierarchy h WHERE $rolid = h.child AND h.parent = l.rolid ORDER BY l.rolid, length (l.url)");
		if ($dbh->errstr) {
			$err = '    : ' . $dbh->errstr;
			exit;
		}
		push @{$lists->[$_->[0]]}, $_->[1] for @$res;

		# Read permissions
		$res = $dbh->selectall_arrayref ("SELECT p.action, p.cat FROM perms p, hierarchy h WHERE $rolid = h.child AND h.parent = p.rolid ORDER BY p.rolid, p.cat");
		if ($dbh->errstr) {
			$err = '    : ' . $dbh->errstr;
			exit;
		}
		push @{$perms->[$_->[0]]}, $_->[1] for @$res;

		print
		start_form ('post', url (-relative => 1)),
		start_div ({-align => 'left'}),
		font ({-size => '+1'}, '  ', b (param ('value'))), hr,
		font ({-size => '+1'}, ' :'), br,
		map ($_ . br, @{$lists->[1]}), hr,
		font ({-size => '+1'}, ' :'), br,
		map ($_ . br, @{$lists->[0]}), hr,
		font ({-size => '+1'}, ' :'), br,
		map ($_ . br, @{$perms->[2]}), hr,
		font ({-size => '+1'}, ' :'), br,
		map ($_ . br, @{$perms->[1]}), hr,
		font ({-size => '+1'}, ' :'), br,
		map ($_ . br, @{$perms->[0]}),
		end_div,
		hidden ('action'),
		hidden ('value'),
		end_form;
	} elsif (param ('action') eq 'delete') {
		if ($rights ne $roltype) {
			$err = '      ';
			exit;
		}

		# Select childs
		$res = $dbh->selectrow_arrayref ("SELECT count(*) FROM hierarchy h WHERE $rolid = h.parent AND h.child != $rolid");
		if ($dbh->errstr) {
			$err = '    : ' . $dbh->errstr;
			exit;
		}
		if ($res->[0]) {
			$err = '    , ..     ';
			exit;
		}

		# Read users
		$res = $dbh->selectrow_arrayref ("SELECT count(*) FROM users u WHERE $rolid = u.rolid");
		if ($dbh->errstr) {
			$err = '    : ' . $dbh->errstr;
			exit;
		}
		if ($res->[0]) {
			$err = '    , ..   ,   ';
			exit;
		}

		if (param ('save')) {
			$dbh->do ("DELETE FROM lists WHERE rolid = $rolid");
			if ($dbh->errstr) {
				$err = '    : ' . $dbh->errstr;
			} else {
				$dbh->do ("DELETE FROM perms WHERE rolid = $rolid");
				if ($dbh->errstr) {
					$err = '    : ' . $dbh->errstr;
				} else {
					$dbh->do ("DELETE FROM hierarchy WHERE child = $rolid");
					if ($dbh->errstr) {
						$err = '    : ' . $dbh->errstr;
					} else {
						$dbh->do ("DELETE FROM roles WHERE rolid = $rolid");
						$err = '    : ' . $dbh->errstr if $dbh->errstr;
					}
				}
			}
		}

		if (param ('save') && !$err) {
			$dbh->commit;
			print font ({-size => '+1'}, '', b (param ('value')), ' ̣');
		} else {
			print font ({-size => '+1', -color => 'red'}, $err), hr if $err;
			undef $err;
			print
			start_form ('post', url (-relative => 1)),
			table ({-border => 1},
				caption (font ({-size => '+1'}, ' ', b (param ('value')))),
				TR (th ('   , ', submit ('save', '')))
			),
			hidden ('action'),
			hidden ('value'),
			end_form;
		}
	} elsif (param ('action') eq 'edit') {
#print start_div ({-align => 'left'});
#for (param ()) {
#	print "param $_ -> '", b (param ($_)), "'", br;
#}
#print end_div;

		$res = $dbh->selectall_arrayref ("SELECT rolid, rname FROM roles WHERE rolid != $rolid" . ($rights eq 'g' ? " AND type = 'g'" : ''));
		if ($dbh->errstr) {
			$err = '    : ' . $dbh->errstr;
			exit;
		}
		my $rolids = {map @$_, @$res};

		my ($cmd, $arg) = map defined param ($_) ? param ($_) : '', qw(cmd arg);
		if (!$cmd) {
			# Do nothing
		} elsif (!grep $cmd eq $_, $rights eq 'g' ? ('allow') : (), qw (parents lstdel prmdel black white redirect reject)) {
			$err = 'a ';
		} elsif ($cmd eq 'parents') {
			if ($rights ne 'g' && !param ('parents')) {
				$err = '   ';
			} elsif (grep !$rolids->{$_}, param ('parents')) {
				$err = '  ';
			} else {
				$dbh->do ("DELETE FROM hierarchy WHERE child = $rolid");
				if ($dbh->errstr) {
					$err = '    : ' . $dbh->errstr;
				} else {
					for ($rolid, param ('parents')) {
						$dbh->do ("INSERT INTO hierarchy VALUES ($_, $rolid)");
						if ($dbh->errstr) {
							$err = '    : ' . $dbh->errstr;
							last;
						}
						param ('parents', '');
					}
				}
			}
		} elsif ($cmd eq 'lstdel') {		# Check URL delete
			if (!$arg) {
				$err = '  URL  ';
			} elsif ($arg !~ /^\d+$/) {
				$err = ' URL  ';
			} else {
				$dbh->do ("DELETE FROM lists WHERE lstid = $arg AND rolid = $rolid");
				$err = '    : ' . $dbh->errstr if $dbh->errstr;
			}
		} elsif	($cmd eq 'prmdel') {		# Check category delete
			if (!$arg) {
				$err = ' a   ';
			} elsif ($arg !~ /^\d+$/) {
				$err = '   ';
			} else {
				$dbh->do ("DELETE FROM perms WHERE prmid = $arg AND rolid = $rolid");
				$err = '    : ' . $dbh->errstr if $dbh->errstr;
			}
		} elsif (grep $cmd eq $_, qw(black white)) {	# Check URL add
			my ($action, $url, $exact) = ($cmd eq 'black' ? 1 : 0, map param ("$cmd-$_"), qw(url exact));
			if (!$url) {
				$err = '  URL';
			} elsif ($url !~ /@{[URLPattern]}/) {
				$err = ' URL ' . b ($url);
			} else {
				if (!$exact) {
					$url .= '*';
				} elsif (!length $2) {
					$url .= '/';
				}
				my $stars = $url =~ /\*/g;
				$stars = 0 if !$stars;
				$url =~ s/([?.+|\[\](){}])/\\$1/g;
				$url =~ s/\*/\.*/g;
				$dbh->do ("INSERT INTO lists (rolid, url, stars, action) VALUES ($rolid, " . $dbh->quote ("^$url\$") . ", $stars, $action)");
				if ($dbh->errstr) {
					$err = '    : ' . $dbh->errstr;
				} else {
					param ("$cmd-$_", '') for qw(url exact);
				}
			}
		} elsif (grep $cmd eq $_, $rights eq 'g' ? ('allow') : (), qw(redirect reject)) {	# Check category add
			my ($action, $cat) = ($cmd eq 'redirect' ? 2 : $cmd eq 'reject' ? 1 : 0, param ("$cmd-cat"));
			if (!$cat) {
				$err = '  ';
			} else {
				my $cats = [grep length $_, split /[\s,:]/, $cat];
				if (!@$cats) {
					$err = '  ';
				} elsif (grep !/^\d+$/, @$cats) {
					$err = ' : ' . join (' ', map b ($_), grep !/^\d+$/, @$cats);
				} else {
					for (@$cats) {
						$dbh->do ("INSERT INTO perms (rolid, cat, action) VALUES ($rolid, $_, $action)");
						if ($dbh->errstr) {
							$err = '    : ' . $dbh->errstr;
							last;
						}
						param ("$cmd-cat", '') if !$err;
					}
				}
			}
		}

		if ($cmd && !$err) {
			$dbh->commit;
			print font ({-size => '+1'}, '  '), hr;
		} else {
			print font ({-size => '+1', -color => 'red'}, $err), hr if $err;
			undef $err;
		}

		my ($lists, $perms, $parents) = ([[], []], [[], [], []]);

		# Read parents
		$res = $dbh->selectall_arrayref ("SELECT parent FROM hierarchy WHERE child = $rolid AND parent != $rolid");
		if ($dbh->errstr) {
			$err = '    : ' . $dbh->errstr;
			exit;
		}
		$parents = [map $_->[0], @$res];

		# Read lists
		$res = $dbh->selectall_arrayref ("SELECT action, url, lstid FROM lists l WHERE l.rolid = $rolid ORDER BY length (url)");
		if ($dbh->errstr) {
			$err = '    : ' . $dbh->errstr;
			exit;
		}
		push @{$lists->[$_->[0]]}, [@$_ [1, 2]] for @$res;


		# Read permissions
		$res = $dbh->selectall_arrayref ("SELECT action, cat, prmid FROM perms p WHERE rolid = $rolid ORDER BY cat");
		if ($dbh->errstr) {
			$err = '    : ' . $dbh->errstr;
			exit;
		}
		push @{$perms->[$_->[0]]}, [@$_ [1, 2]] for @$res;

		param ($_, '') for qw(cmd arg);

		print
		start_form ('post', url (-relative => 1)),
		font ({-size => '+1'}, ' ', b (param ('value'))), br, br,
		table ({-border => 1},
			caption (font ({-size => '+1'}, '')),
			TR (td (checkbox_group (
				-name	=> 'parents',
				-values	=> [sort {$a <=> $b} keys %$rolids],
				-defaults=> $parents,
				-labels	=> $rolids,
				-cols	=> 3,
				-override=> 1))),
			TR (th ({-align => 'center'}, button (undef, '', "cmd.value='parents'; submit ()")))
		), br,
		table ({-border => 1},
			caption (font ({-size => '+1'}, ' ')),
			map (TR (
				th ({-align => 'left'}, $_->[0]),
				td (button (undef, '', "cmd.value='lstdel'; arg.value=$_->[1]; submit ()"))
			), @{$lists->[1]}),
			TR (
				td ({-align => 'left'}, textfield ('black-url', undef, 40, 100)),
				td (checkbox ('black-exact', undef, 1, ' '), br, button (undef, '', "cmd.value='black'; submit()"))
			)
		), br,
		table ({-border => 1},
			caption (font ({-size => '+1'}, ' ')),
			map (TR (
				th ({-align => 'left'}, $_->[0]),
				td (button (undef, '', "cmd.value='lstdel'; arg.value=$_->[1]; submit()"))
			), @{$lists->[0]}),
			TR (
				td ({-align => 'left'}, textfield ('white-url', undef, 40, 100)),
				td (checkbox ('white-exact', undef, 1, ' '), br, button (undef, '', "cmd.value='white'; submit()"))
			)
		), br,
		table ({-border => 1},
			caption (font ({-size => '+1'}, ' ')),
			map (TR (
				th ({-align => 'left'}, $_->[0]),
				td (button ('delprm', '', "cmd.value='prmdel'; arg.value=$_->[1]; submit()"))
			), @{$perms->[2]}),
			TR (
				td ({-align => 'left'}, textfield ('redirect-cat', undef, 40, 100)),
				td (button (undef, '', "cmd.value='redirect'; submit()"))
			)
		), br,
		table ({-border => 1},
			caption (font ({-size => '+1'}, ' ')),
			map (TR (
				th ({-align => 'left'}, $_->[0]),
				td (button ('delprm', '', "cmd.value='prmdel'; arg.value=$_->[1]; submit()"))
			), @{$perms->[1]}),
			TR (
				td ({-align => 'left'}, textfield ('reject-cat', undef, 40, 100)),
				td (button (undef, '', "cmd.value='reject'; submit()"))
			)
		), br,
		$rights eq 'g' ? (table ({-border => 1},
			caption (font ({-size => '+1'}, ' ')),
			map (TR (
				th ({-align => 'left'}, $_->[0]),
				td (button ('delprm', '', "cmd.value='prmdel'; arg.value=$_->[1]; submit()"))
			), @{$perms->[0]}),
			TR (
				td ({-align => 'left'}, textfield ('allow-cat', undef, 40, 100)),
				td (button (undef, '', "cmd.value='allow'; submit()"))
			)
		), br) : '',
		map (hidden ($_), qw(action value cmd arg)),
		end_form;
	}
}
