#!/usr/bin/perl -w

use strict;
use Digest::MD5 qw(md5_base64);
use CGI qw(header *html *div font table caption TR th td a b hr cookie remote_addr user_agent param url *form textfield popup_menu submit hidden);
use DBI;
require './config.ph';

if (!$::DSN) {
	print
	header (-charset => 'koi8-r'),
	start_html,
	font ({-size => '+1', -color => 'red'}, '     .   config.ph'),
	end_html;
	exit;
}

use constant	UserNamePattern => '^[\w\.]+$';

if (grep (defined $_, map cookie ($_), qw(aname auth)) != 2 ) {	# Check cookie
	print header (-location => 'login.cgi');
	exit;
}

my ($err, $rights, $res);
my $dbh = DBI->connect ($::DSN, $::DBUser, $::DBPass, {ShowErrorStatement => 1, AutoCommit => 0});
if (!$dbh) {
	$err = '    : ' . DBI->errstr;
} else {
	my $u = $dbh->selectrow_arrayref ("SELECT passwd, cookie, type FROM admins WHERE aname = " . $dbh->quote (cookie ('aname')));
	if ($dbh->errstr) {
		$err = '    : ' . $dbh->errstr;
	} elsif ($u && cookie ('auth') eq md5_base64 ($u->[1], remote_addr, user_agent, cookie ('aname'), $u->[0])) {
		$rights = $u->[2];
	} else {
		if ($u) {
			$dbh->do ("UPDATE admins SET cookie = '' WHERE aname = " . $dbh->quote (cookie ('aname')));
			$dbh->commit;
		}
		print header (-location => 'login.cgi');
		exit;
	}
}

print
header (-charset => 'koi8-r'),
start_html (-title => ' ', -encoding => 'koi8-r'),
start_div ({-align => 'center'});
END {
	$dbh->disconnect if $dbh;
	print
	$err ? font ({-size => '+1', -color => 'red'}, $err) : '',
	hr,
	a ({-href => 'main.cgi'}, '   '),
	end_div,
	end_html;
}

exit if $err;

if ($rights eq 'g') {
	$err = '      ';
	exit;
}

if (!defined param ('action') || !grep param ('action') eq $_, qw(new edit delete)) {
	$err = 'oe ';
	exit;
}

if (param ('action') eq 'edit' || param ('action') eq 'delete') {
	if (!param ('value') || param ('value') !~ /^\d+$/) {
		$err = '  ';
		exit;
	}

	$res = $dbh->selectrow_arrayref ("SELECT count(useid) FROM users WHERE useid = " . param ('value'));
	if ($dbh->errstr) {
		$err = '    : ' . $dbh->errstr;
		exit;
	}
	if (!$res->[0]) {
		$err = '   ';
		exit;
	}
}

if (param ('action') eq 'new' || param ('action') eq 'edit') {

	# Read users
	my ($users, $u);
	$res = $dbh->selectall_arrayref ("SELECT useid, uname, addr, mask, rolid FROM users");
	if ($dbh->errstr) {
		$err = '    : ' . $dbh->errstr;
		exit;
	}
	for (@$res) {
		if ($_->[0] == param ('value')) {
			$u = {useid => $_->[0], uname => $_->[1], addr => $_->[2], mask => $_->[3], rolid => $_->[4]};
		} else {
			push @$users, $_;
		}
	}

	# Read roles
	$res = $dbh->selectall_arrayref ("SELECT rolid, rname FROM roles");
	if ($dbh->errstr) {
		$err = '    : ' . $dbh->errstr;
		exit;
	}
	my $rolids = {map @$_, @$res};

	if (param ('save')) {
		if (param ('uname') && param ('uname') !~ /@{[UserNamePattern]}/) {
			$err = '  ';
		} elsif (!param ('addr')) {
			$err = '  IP ';
		} elsif (param ('addr') !~ /^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})$/ || $1 > 255 || $2 > 255 || $3 > 255 || $4 > 255) {
			$err = ' IP  ' . b (param ('addr'));
		} elsif (param ('mask') && (param ('mask') !~ /^((\/(\d{1,2}))|(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3}))$/ || $2 && $3 > 32 || !$2 && ($4 > 255 || $5 > 255 || $6 > 255 || $7 > 255))) {
			$err = ' /  ' . b (param ('mask'));
		} elsif (!param ('rolid')) {
			$err = '  ';
		} elsif (!$rolids->{param ('rolid')}) {
			$err = '  ';
		} else {
			$u->{uname} = param ('uname');
			$u->{mask} = !param ('mask') ? 0xffffffff : $2 ? 0xffffffff << (32 - $3) : unpack "N", pack "C4", $4, $5, $6, $7;
			$u->{addr} = unpack "N", pack "C4", split /\./, param ('addr');
			$u->{mask} = 0 if !$u->{addr};
			$u->{addr} &= $u->{mask};
			$u->{rolid} = param ('rolid');
			if (grep $u->{uname} eq $_->[1] && $u->{addr} == $_->[2] && $u->{mask} == $_->[3], @$users) {
				$err = '   IP   ';
			} else {
				$dbh->do (param ('action') eq 'new'
					? "INSERT INTO users (uname, addr, mask, rolid) VALUES (" . $dbh->quote ($u->{uname}) . ", $u->{addr}, $u->{mask}, $u->{rolid})"
					: "UPDATE users SET uname = " . $dbh->quote ($u->{uname}) . ", addr = $u->{addr}, mask = $u->{mask}, rolid = $u->{rolid} WHERE useid = $u->{useid}"
				);
				$err = '    : ' . $dbh->errstr if $dbh->errstr;
			}
		}
	} elsif (param ('action') eq 'edit') {
		if (!$u->{mask}) {
			$u->{addr} = '0.0.0.0';
			$u->{mask} = '';
		} else {
			$u->{addr} = join '.', unpack "C4", pack "N", $u->{addr};
			my ($bits, $mask);
			for ($bits = 32, $mask = 0xffffffff; $bits >= 0 && $u->{mask} != $mask; $bits--, $mask <<= 1)
				{}
			if ($bits < 0) {
				$u->{mask} = join '.', unpack "C4", pack "N", $u->{mask};
			} elsif ($bits != 32) {
				$u->{mask} = "/$bits";
			} else {
				$u->{mask} = '';
			}
		}
		param ($_, $u->{$_}) for qw(uname addr mask rolid);
	}

	if (param ('save') && !$err) {
		$dbh->commit;
		print font ({-size => '+1'}, '  ', param ('action') eq 'new' ? '' : '');
	} else {
		print font ({-size => '+1', -color => 'red'}, $err), hr if $err;
		undef $err;
		print
		start_form ('post', url (-relative => 1)),
		table ({-border => 1},
			caption (font ({-size => '+1'}, ' ')),
			TR (th (' '), td (textfield ('uname', undef, 20, 20))),
			TR (th ('IP '), td (textfield ('addr', undef, 20, 20))),
			TR (th (' /'), td (textfield ('mask', undef, 20, 20))),
			TR (th (''), td (popup_menu ('rolid', [sort {$a <=> $b} keys %$rolids], undef, $rolids))),
			TR (th ({-colspan => 2, align => 'center'}, submit ('save', param ('action') eq 'new' ? '' : '')))
		),
		hidden ('action'),
		hidden ('value'),
		end_form;
	}
} elsif (param ('action') eq 'delete') {
	if (param ('save')) {
		$dbh->do ("DELETE FROM users WHERE useid = " . param ('value'));
		$err = '    : ' . $dbh->errstr if $dbh->errstr;
	}

	if (param ('save') && !$err) {
		$dbh->commit;
		print font ({-size => '+1'}, '  ̣');
	} else {
		print font ({-size => '+1', -color => 'red'}, $err), hr if $err;
		undef $err;
		print
		start_form ('post', url (-relative => 1)),
		table ({-border => 1},
			caption (font ({-size => '+1'}, ' ')),
			TR (th ('   , ', submit ('save', '')))
		),
		hidden ('action'),
		hidden ('value'),
		end_form;
	}
}
